'''
data_helpers: data manipulation utilities
'''

import dateparser
import datetime
from   datetime import datetime as dt
from   dateutil import tz


# Constants.
# .............................................................................

DATE_FORMAT = '%b %d %Y %H:%M:%S %Z'
'''Format in which lastmod date is printed back to the user. The value is used
with datetime.strftime().'''


# Functions.
# .............................................................................

def slice(lst, n):
    # Original algorithm from Jurgen Strydom posted 2019-02-21 Stack Overflow
    # https://stackoverflow.com/a/54802737/743730
    '''Yield n number of slices from lst.'''
    for i in range(0, n):
        yield lst[i::n]


def ordinal(n):
    '''Print a number followed by "st" or "nd" or "rd", as appropriate.'''
    # Spectacular algorithm by user "Gareth" at this posting:
    # http://codegolf.stackexchange.com/a/4712
    return '{}{}'.format(n, 'tsnrhtdd'[(n/10%10!=1)*(n%10<4)*n%10::4])


def expand_range(text):
    '''Return individual numbers for a range expressed as X-Y.'''
    # This makes the range 1-100 be 1, 2, ..., 100 instead of 1, 2, ..., 99
    if '-' in text:
        range_list = text.split('-')
        range_list.sort(key = int)
        return [*map(str, range(int(range_list[0]), int(range_list[1]) + 1))]
    else:
        return text


def timestamp():
    '''Return a string describing the date and time right now.'''
    return dt.now(tz = tz.tzlocal()).strftime(DATE_FORMAT)


def parse_datetime(string):
    '''Parse a human-written time/date string using dateparser's parse()
function with predefined settings.'''
    return dateparser.parse(string, settings = {'RETURN_AS_TIMEZONE_AWARE': True})


def plural(word, count):
    '''Simple pluralization; adds "s" to the end of "word" if count > 1.'''
    if isinstance(count, int):
        return word + 's' if count > 1 else word
    elif isinstance(count, (list, set, dict)):
        return word + 's' if len(count) > 1 else word
    else:
        # If we don't recognize the kind of thing it is, return it unchanged.
        return word



'''
interruptions.py: provide an interruptible wait(...) and related utilities.

This module includes wait(...), a replacement for sleep(...) that is
interruptible and works with multiple threads.  It also provides methods to
cause an interruption, check whether an interruption occurred, and other
related operations.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code is
open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import threading

from .debug import log
from .exceptions import *


# Global variables.
# .............................................................................

__waiter = threading.Event()
'''Internal state variable used to record if an interrupt occurred.'''


# Main functions.
# .............................................................................

def wait(duration):
    '''Wait for "duration" seconds, in a way that can be interrupted.

    This is a replacement for sleep(duration).  If interrupted, this function
    raises the exception UserCancelled.
    '''
    if __debug__: log(f'waiting for {duration} s')
    __waiter.wait(duration)
    if interrupted():
        if __debug__: log(f'raising UserCancelled')
        raise UserCancelled('Interrupted while waiting')


def interrupt():
    '''Interrupt any waits and internally record an interrupt has occurred.'''
    if __debug__: log(f'interrupting wait')
    __waiter.set()


def interrupted():
    '''Return True if interrupt() has been called and not cleared.'''
    return __waiter.is_set()


def raise_for_interrupts():
    '''Check whether an interrupt occurred; if so, raise UserCancelled.'''
    if interrupted():
        if __debug__: log(f'raising UserCancelled')
        raise UserCancelled('Interrupted while waiting')


def reset():
    '''Clear the internal marker that an interrupt occurred.'''
    if __debug__: log(f'clearing wait')
    __waiter.clear()



'''
network.py: miscellaneous network utilities for Microarchiver

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2018-2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import http.client
from   http.client import responses as http_responses
from   os import stat
import requests
from   requests.packages.urllib3.exceptions import InsecureRequestWarning
import socket
import ssl
import urllib
from   urllib import request
import urllib3
import warnings

from .debug import log
from .exceptions import *
from .interruptions import wait, interrupted


# Constants.
# .............................................................................

_MAX_RECURSIVE_CALLS = 10
'''How many times can certain network functions call themselves upon
encountering a network error before they stop and give up.'''

_MAX_FAILURES = 3
'''Maximum number of consecutive failures before pause and try another round.'''

_MAX_RETRIES = 5
'''Maximum number of times we back off and try again.  This also affects the
maximum wait time that will be reached after repeated retries.'''


# Main functions.
# .............................................................................

def network_available(address = "8.8.8.8", port = 53, timeout = 5):
    '''Return True if it appears we have a network connection, False if not.
    By default, this attempts to contact one of the Google DNS servers (as a
    plain TCP connection, not as an actual DNS lookup).  Argument 'address'
    and 'port' can be used to test a different server address and port.  The
    socket connection is attempted for 'timeout' seconds.
    '''
    # Portions of this code are based on the answer by user "7h3rAm" posted to
    # Stack Overflow here: https://stackoverflow.com/a/33117579/743730
    try:
        if __debug__: log('testing if we have a network connection')
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((address, port))
        if __debug__: log('we have a network connection')
        return True
    except Exception:
        if __debug__: log('could not connect to https://www.google.com')
        return False


def hostname(url):
    parsed = urllib.parse.urlsplit(url)
    return parsed.hostname


def scheme(url):
    parsed = urllib.parse.urlsplit(url)
    return parsed.scheme


def netloc(url):
    parsed = urllib.parse.urlsplit(url)
    return parsed.netloc


def timed_request(method, url, session = None, timeout = 20, **kwargs):
    '''Perform a network access, automatically retrying if exceptions occur.

    The value given to parameter "method" must be a string chosen from among
    valid HTTP methods, such as "get", "post", or "head".  If "session" is
    not None, it is used as a requests.Session object. "Timeout" is a timeout
    (in seconds) on the network requests get or post. Other keyword arguments
    are passed to the network call.
    '''
    failures = 0
    retries = 0
    error = None
    while failures < _MAX_FAILURES and not interrupted():
        try:
            with warnings.catch_warnings():
                # The underlying urllib3 library used by the Python requests
                # module will issue a warning about missing SSL certificates.
                # We don't care here.  See also this for a discussion:
                # https://github.com/kennethreitz/requests/issues/2214
                warnings.simplefilter("ignore", InsecureRequestWarning)
                if __debug__: log(f'doing http {method} on {url}')
                func = getattr(session, method) if session else getattr(requests, method)
                response = func(url, timeout = timeout, verify = False, **kwargs)
                if __debug__: log(f'response received: {response}')
                return response
        except (KeyboardInterrupt, UserCancelled) as ex:
            if __debug__: log('received UserCancelled during network operation')
            raise
        except Exception as ex:
            if ex.args and len(ex.args) > 0:
                if isinstance(ex.args[0], urllib3.exceptions.MaxRetryError):
                    # No point in retrying if we get this.
                    raise
            # Problem might be transient.  Don't quit right away.
            failures += 1
            if __debug__: log(f'exception (failure #{failures}): {str(ex)}')
            # Record the first error we get, not the subsequent ones, because
            # in the case of network outages, the subsequent ones will be
            # about being unable to reconnect and not the original problem.
            if not error:
                error = ex
            # Pause briefly b/c it's rarely a good idea to retry immediately.
            if __debug__: log('pausing for 1 s')
            wait(0.5)
        if failures >= _MAX_FAILURES:
            # Pause with exponential back-off, reset failure count & try again.
            if retries < _MAX_RETRIES:
                retries += 1
                failures = 0
                pause = 10 * retries * retries
                if __debug__: log(f'pausing {pause}s due to consecutive failures')
                wait(pause)
            else:
                if __debug__: log('exceeded max failures and max retries')
                raise error
    if interrupted():
        if __debug__: log('interrupted -- raising UserCancelled')
        raise UserCancelled('Network request has been interrupted')
    else:
        # In theory, we should never reach this point.  If we do, then:
        raise InternalError('Unexpected code contingency in timed_request')


def net(method, url, session = None, timeout = 20, handle_rate = True,
        polling = False, recursing = 0, **kwargs):
    '''Invoke HTTP "method" on 'url' with optional keyword arguments provided.

    Returns a tuple of (response, exception), where the first element is
    the response from the get or post http call, and the second element is
    an exception object if an exception occurred.  If no exception occurred,
    the second element will be None.  This allows the caller to inspect the
    response even in cases where exceptions are raised.

    If keyword 'session' is not None, it's assumed to be a Python requests
    Session object to use for the network call.

    If keyword 'handle_rate' is True, this function will automatically pause
    and retry if it receives an HTTP code 429 ("too many requests") from the
    server.  If False, it will return the exception RateLimitExceeded instead.

    If keyword 'polling' is True, certain statuses like 404 are ignored and
    the response is returned; otherwise, they are considered errors.  The
    behavior when True is useful in situations where a URL does not exist until
    something is ready at the server, and the caller is repeatedly checking
    the URL.  It is up to the caller to implement the polling schedule and
    call this function (with polling = True) as needed.

    This method always passes the argument allow_redirects = True to the
    underlying Python requests library network call.
    '''
    def addurl(text):
        return f'{text} for {url}'

    resp = None
    try:
        resp = timed_request(method, url, session, timeout,
                             allow_redirects = True, **kwargs)
    except requests.exceptions.ConnectionError as ex:
        if __debug__: log(f'got network exception: {str(ex)}')
        if recursing >= _MAX_RECURSIVE_CALLS:
            if __debug__: log('returning NetworkFailure')
            return (resp, NetworkFailure(addurl('Too many connection errors')))
        arg0 = ex.args[0]
        if isinstance(arg0, urllib3.exceptions.MaxRetryError):
            if __debug__: log(str(arg0))
            # This can be a transient error due to various causes.  Retry once.
            if recursing == 0:
                wait(1)
                return net(method, url, session, timeout, handle_rate,
                           polling, recursing + 1, **kwargs)
            else:
                # We've seen maxretry before or we're recursing due to some
                # other failure.  Time to bail.
                original = unwrapped_urllib3_exception(arg0)
                if __debug__: log('returning NetworkFailure')
                if isinstance(original, str) and 'unreacheable' in original:
                    return (resp, NetworkFailure(addurl('Unable to connect to server')))
                elif network_available():
                    return (resp, NetworkFailure(addurl('Unable to resolve host')))
                else:
                    return (resp, NetworkFailure(addurl('Lost connection with server')))
        elif (isinstance(arg0, urllib3.exceptions.ProtocolError)
              and arg0.args and isinstance(args0.args[1], ConnectionResetError)):
            if __debug__: log('net() got ConnectionResetError; pausing for 1 s')
            wait(1)                     # Sleep a short time and try again.
            if __debug__: log(f'doing recursive call #{recursing + 1}')
            return net(method, url, session, timeout, polling, handle_rate,
                       recursing + 1, **kwargs)
        else:
            if __debug__: log('returning NetworkFailure')
            return (resp, NetworkFailure(str(ex)))
    except requests.exceptions.ReadTimeout as ex:
        if network_available():
            if __debug__: log('returning ServiceFailure')
            return (resp, ServiceFailure(addurl('Timed out reading data from server')))
        else:
            if __debug__: log('returning NetworkFailure')
            return (resp, NetworkFailure(addurl('Timed out reading data over network')))
    except requests.exceptions.InvalidSchema as ex:
        if __debug__: log('returning NetworkFailure')
        return (resp, NetworkFailure(addurl('Unsupported network protocol')))
    except Exception as ex:
        if __debug__: log(f'returning exception: {str(ex)}')
        return (resp, ex)

    # Interpret the response.  Note that the requests library handles code 301
    # and 302 redirects automatically, so we don't need to do it here.
    error = None
    code = resp.status_code
    if __debug__: log(addurl(f'got http status code {code}'))
    if code == 400:
        error = RequestError(addurl('Server rejected the request'))
    elif code in [401, 402, 403, 407, 451, 511]:
        error = AuthenticationFailure(addurl('Access is forbidden'))
    elif code in [404, 410] and not polling:
        error = NoContent(addurl("No content found"))
    elif code in [405, 406, 409, 411, 412, 414, 417, 428, 431, 505, 510]:
        error = InternalError(addurl(f'Server returned code {code} ({resp.reason})'))
    elif code in [415, 416]:
        error = ServiceFailure(addurl('Server rejected the request ({resp.reason})'))
    elif code == 429:
        if handle_rate and recursing < _MAX_RECURSIVE_CALLS:
            pause = 5 * (recursing + 1)   # +1 b/c we start with recursing = 0.
            if __debug__: log(f'rate limit hit -- sleeping {pause} s')
            wait(pause)                   # 5 s, then 10 s, then 15 s, etc.
            if __debug__: log(f'doing recursive call #{recursing + 1}')
            return net(method, url, session = session, timeout = timeout,
                       polling = polling, handle_rate = True,
                       recursing = recursing + 1, **kwargs)
        error = RateLimitExceeded('Server blocking requests due to rate limits')
    elif code == 503:
        error = ServiceFailure(f'{resp.reason}')
    elif code == 504:
        error = ServiceFailure(f'Server timeout: {resp.reason}')
    elif code in [500, 501, 502, 506, 507, 508]:
        error = ServiceFailure(addurl(f'Server error (code {code} -- {resp.reason})'))
    elif not (200 <= code < 400):
        error = NetworkFailure(f'Unable to resolve {url}')
    if __debug__: log('returning result {}',
                      f'with error {error}' if error else 'without error')
    return (resp, error)


# Next code originally from https://stackoverflow.com/a/39779918/743730
def disable_ssl_cert_check():
    requests.packages.urllib3.disable_warnings()
    try:
        _create_unverified_https_context = ssl._create_unverified_context
    except AttributeError:
        # Legacy Python that doesn't verify HTTPS certificates by default
        pass
    else:
        # Handle target environment that doesn't support HTTPS verification
        ssl._create_default_https_context = _create_unverified_https_context


def unwrapped_urllib3_exception(ex):
    if hasattr(ex, 'args') and isinstance(ex.args, tuple):
        return unwrapped_urllib3_exception(ex.args[0])
    else:
        return ex



'''
logo.py: Embedded icon for use in the "About" panel

The rest of this file was generated by using the img2py.py program that
is provided with wxPython 4.0, with the following commands:

  inkscape -o eprints2archives-icon.png  -w 160  ../.graphics/eprints2archives-icon.svg
  img2py -f -i -n Logo eprints2archives-icon.png logo.py
  rm eprints2archives-icon.png

The only changes to the output were to add the file header and this
explanatory comment.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code is
open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

#----------------------------------------------------------------------
# This file was generated by /usr/local/bin/img2py
#
from wx.lib.embeddedimage import PyEmbeddedImage

Logo = PyEmbeddedImage(
    b'iVBORw0KGgoAAAANSUhEUgAAAKAAAADJCAYAAACkNDm/AAAACXBIWXMAABqKAAAaigFGPbT9'
    b'AAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAHGtJREFUeJztnXl8VOW5'
    b'x7/PORMSkDVIZUkmQahawWqLqF1cWFSsV2u12GrVWr1FBSaBLtZumrbXi6W1YTKASovaxRVb'
    b'e6sFLSpXtFw3atVaRRGYmbCJgCwKJDPnuX9MQJZMcs7MnDkzk/P9fPL5wORdfjP5zTnved/n'
    b'fV7BxxP0W4OrscxRqHwKi6MRqoEgcFjbTwDY1vazE2UtIq8DbyDJ1+jXvFwaSHj3DnKDeC2g'
    b'q6ATMRkYHIPoBWCcCXpUlk1uB30SlcdRfURmx9flRGie8Q3oMjo9eCQJnYzIpcAgl7pJoCwE'
    b'mceG6GOygKRL/eQc34AuodNqT0CTP0bli4CZx67XIPJf9Iv+thhu0b4Bc4xOq63F0hmgFwOG'
    b'd0rkLbBuoin+gIB6p6NjfAPmCG0gwJbgFOC/gJ5e69mPpZjW1dLYvNJrIe3hGzAHaKh6GGLc'
    b'A3qy11rSsBPR7xGO31ZoV0PfgFmi9cGJKL8BenutpXPkfna3Xi3z1n3otZK9+AbMEAUhFLwe'
    b'4b/xdKznFH0Fy/iSzI6u9loJ+AbMCJ2IyaDgfODrXmvJkA2YjJfG2OteC/EN6BCdNKqMik33'
    b'ABNz0Nwe4EVgGfAv1HibgPUuRmI7HwaSGGZvzEQvLKMaYSSixwHjgMFZ9yy8S9IaL7ObX8u6'
    b'raxk+NhGQair/gPIpVk0k0T5K6L3kdj9qMzdtNOxjtDQ4zGSF6L8J9mZ8T0sa6yXJvQN6ACt'
    b'q7kV9FsZVt+NcDtihGXWmjU50dNAgK01XwS9CeW4DFuJkTBPlrlrNuRCk1N8A9pE62uuRvU3'
    b'GVUWuY+E8X2ZszqaY1kAaAMGW4OXodxCRst98jxmcow0Nu/KubjOes53h8WI1lePRmUpUOGw'
    b'6jqEayQce9QNXQej06sqSUokwyHC76UpdkXORXWCb8BO0MkDehLo/k9gmLOa8jQtgYvl9nfe'
    b'dUVYB2h98AqUeUC5w5qXSFP8fldEpaGkDKgNBNg+eBB7pDum2QtDPsRo3UnvdeszXZjX+uDt'
    b'KNc4rDWfyvi1XgYDaF3NZxF9GOVjDqptxbSOk8bmta4JO4iiNqBOrT0GU8ejehrI8aBDgbJ2'
    b'iiaAKMpLwHNYyYUyZ+1bnbYfqjodMZbg6HPSX9IUv74Qlrw0FDwW4SngCNuVRBdJOP4F91Qd'
    b'1F2+OsoVOq22L6pXo9blKdNlzBuo3EVr4Lft3SZTT5jBfzh7utQ50hSfmoWmnKPTgyNIyhLQ'
    b'AbYrCefla9xaNAbU6VWVJOQHiEwCeuWw6V3AbzBkhsyKrt/XX331ZFTmOFD4MJXxL0sDVg61'
    b'5QQNVZ+KyBNAN5tV3qQydlw+hhAFb8DUmmv1tYj8DOjvYlc7QH7K+mgjh9WW0dtaCQyxWXcF'
    b'Fd1Hy8wVO1zUlxVaH/xm24OJTWSqNEUdfAEzo6ANqNOrhpA07gLOzF+n/B8Gz6J812aNFsQa'
    b'LeHmV13VlQO0LvggdpcQlTh7BgyTectb3dRUsFEcGqo9haSxnHyaD0D4jAPzgXJzMZgPgLLE'
    b'ZMDeiodQTff3crHe3SEFaUAN1VyIWEtw8vTmCfIW/Xvd4rUKu8it695DnHy5Ml52tE3BGVDr'
    b'gxMRvR/nqw75R/m+NLze4rUMR/SL3YvyD5ulR2mo9hQ35RSUATUUHI9yD+3P5RUWwksSif7J'
    b'axlOkQYsDG6wXcGwvuKinMIxoE6tPQZhAdmZrwXkVUQXAQtQHkd4Dcj9QFr11py3mSckHFsM'
    b'/NNWYWWiNrjnk4BbDTtBr6ytwLQeROmbQfUPQe9D5AGsbkslsnJPu+33SpyMyESQr5L9dM5a'
    b'KuMPZdmGt6iGEbnLRskhbK3+PMSXuiGjIAxIb+uWDOLZWhHmEEjcLLeue6+jgnL3mt3A08DT'
    b'OnnADZg9rsPQH6L0yUivcG8xbPrukB3m/fS2wtjZTGUZZwOuGNDzW7BOq/k04HT56k3gMxKO'
    b'Te/MfAcjczftlEj0FyT1WGChw37bGjHuzaheAdH2pXzEXmE91S0dnhpQQbA0grPUFYvZHThF'
    b'mmLLs+lbZsfXURk7D+HnDquulVlr7I2fCh1D7Q4jTtIra12ZlfD2ClgfPBf4rIMai9lunC/z'
    b'Vm3LRffSgCXh2A2gP3NQ7Ylc9F0QdOvxJNhKZFROH+tENyR4a0Dlxw5K/4uK7he13TpyijTF'
    b'b0T5nb3CYu+2VQTIzBU7UF6xVdhihBsaPDNg2wTnSTaL7yKZvMjVxf7krimphD4d8g7b5K+u'
    b'afAC4e/2yulwN7r37goo1tUOCt9oJ4A0G2Tupp2IcT6Q7qFmK2pd6sYV2FNU37BXzigdA2rD'
    b'iG6I7Y3dq6js2eSqoDYkvHoFlh4PchfC3nHmduAeTOskiTS/kA8deeYdW6UMdbgnxh7ezANu'
    b'3nEaYnsO7uf5XG9tS3V7FXCVXn90L2au2FkI4fWuEdBVJG1E5SmVbnTvzS1YxN6eA2EbuxN/'
    b'cFlN+u5nrthR0uZLscVWKXEn56E3BlS1F2Gh/LmQUomVJH362EsNohzmRveuRURrw4hubHq/'
    b'FtPshdIPkSQG25DkdpLGq0D3ztXJlyUc/aNbGn1SaF2wBTtBIKbVI9fZE3JmQJ1aPRjhLIQz'
    b'QUaR2sid3RgzYQzyKmdJV6Et25fNMbbcj7KYQPLxXO0dzsqA2jCiG1t2XAB8ExhLbm/pG6Up'
    b'NjCH7fm0g4YGDkC6Oc3eYAFLUJ1P//iCbAIzMjJgKiF39eUgNwK1mXbecSf8n0RiTpbp8o5O'
    b'GtyDHmXH05rY7PY8pVvolCFHYZorsmhiNXAL62PzMzmfxPEVS0NVJ7El+DLInbhlPgDD5tOZ'
    b'R+i02jOoCLyNpcswzRVaX/1Hvf7oXO5Xzg8B6ZdlC0OBOxgUfDGT8H3bBtQGAloXnIEYy4CR'
    b'TjtyjFKwT79aVzUBy1rI/skhVS5kz66ndHqVK/NlrqFGrqZXPoVYz2h98CfaYH/sb8uAOmVI'
    b'f7YEFwE3kK9Tf7TwMgwA6NSac8B4mPae4pUTSRiLdcoQNzfQ5xZDcvlFD6DcyJbgIp1Wayu6'
    b'vVMDamhIFWZgGTA+a3lOEP0gr/3ZQEM152Low3S0Y0/4NKb5hH578OH5U5YF78vLwNYctzoe'
    b'y1qm06s6zSzRoQG1/sggYv5vDk52dI4ab+a9zw7QUPDLiD6MvZx7J9AaeFon1xb8U7zcvWY3'
    b'KjNcaPoTJI0lOrW6wxzWaQ2YykKVeAzHiRlzwgeQuM+DfttF64IXI9yHsx17xxKwOv0DFAT9'
    b'o7eCzHWh5Y9jyOMaGp5230m70zBt554tIvvbbhRYyUfrjX2Aj5N6ckqHhfANCcfsBYi6jIaq'
    b'L0Hkd2Q+qb4C0xqXz6SPmZLKhyhXgAwndWh2f1IJmrLdp72Qyth57WUOa9+A9cGfoNyYYWfP'
    b'gN6FBBZKePXGdtufXDsQU78A1uWIfJ59f1x9BeF7Eo4/nmHfOUVDNVch+muyn2BfjWGMzVV2'
    b'/HyiDSO68d72T2JwNsjlwNEZNSR8X8KxQ9KYHGJAra86GTWexfk3/lnU+K5E1jznpJJOq+2L'
    b'ZQ1DkxslsrbZYZ+u0ZbO7HZyt7oTxWSsNMZW5ai9vJNKlVfzhdTxZPpJh9VbMeQUmRU9IC3I'
    b'AQbUiZgMDr7scI/uHpTricQipRK6pKHgNQhzyXm0kMZQxkokbi8ItEBpWwn7HshNOLs9v0Bl'
    b'7DP734oP/IAHBa92aL4tCOMlEmsqHfNVX4dwG66EqkkQZInWVX08923nD2kgIU3xm4FzcDaF'
    b'cxKba648oK29/9Arayvoba0G7E4dvA+Mz3Z/biGhdTXfAf1FHrraiMm4QjgsMFu0LjiK1FZV'
    b'u2lV1rB7wFF7E19+9C3vrZdh33xJVL9SUuYLBb+XJ/MBHIHFUzq1KsPjtQqHlAfkQrAdEVNL'
    b'+btf2/sfA9oGl+h0B/3+TCLxvzkoX9BoXbABIb+JJpWPYZhPan2V08F8wSFN0SUoP7JdwZAp'
    b'+/4JQKhqNHCsvd54jd0D/tuZxMJFQzU3Azd1UORNRE9nfSyABGqAzvaoJBC+j2n1p9uengiT'
    b'gDRh7zoANZ5qy49T3GyI/RKwt2tQOVFDQ4+HvQYU034SQrWudztxdb7Q+uBMRH/QQYlX0JbT'
    b'JBxfKgtISnhVjMrY1+kwU5TOkHDsFmls3iK/3PiBhGO/Bjl7v22eB9MfS5/Q+urR2bwXr5EF'
    b'JBG+bbuCkbwE9o0B9UKb1ZZLU/NjTsUVGgqi9cHGTpKRL8fUsRLZsGn/F6UBC9X056lJ4J5D'
    b'XmqKLsOyziL9E2M/kMU6reYzdvQXKhKOPYvd3DnKuQCG1h8ZxHZgqbp+boTbKAj1wVko0zoo'
    b'thzTOksam9MFxb7fQd12fyeR5hdQGUe6zAtKHyxdrHU1YzpouwjQ2TYLjtTpVUMMrKTd3G8t'
    b'tMqfM5VVCLSdeB5BqUtbSPg72m1sB+bLGIlEX0Y5nfRHJRwG+qjWBcfluu+8sftjC7E7N5gw'
    b'TzbA+pTNppfKbbFcx43lDW3AoK5mPnz0BNYOz1Le/RyJrNzulg6JxP6NZYwB1qUp0gP4i9YH'
    b'83s+So5IPR+IvbV80dEGiN1ZeXtZlAoQnYjJluCdoN/ooNhSErvOycdxWzJ7zZuIORZIFyHT'
    b'A+URra8+320triDWM/bK6VEGgr2sRyKvZaPJK3QiJoOCdwNfT1+IxzGtCTJ3k70sATlAwqtX'
    b'gDUGSBeAUY7KAp1WfUG+NOUMMV62Vc6SWgO7GeMtii6KQxsIMKj6d8BlaQuJLmKHcUGud/zb'
    b'QZqa3yZpfh7SfrbdsORBDdXYnaUoDHYH7AVbCIMNsJl0xjIKepvkwaQ2zVc/CHJp2kLCX7HK'
    b'v+Rlzj+ZszqKBMaQPk1aGaIPan0w/Zeo0Bj4znvYC07paWAnRwtAgKJJzNiWseFBkC91UOwh'
    b'dg34UnvniuQbCa+KEUiMAd5OU8REuVvrg1fkU1emtIVb2fFLDwO7pwiZe+weduwpGhpezpYd'
    b'DwFfTF+IB6iMXVJIKzryq3VxEsZpKOkiZEyUO7WuuqMHqULCTpxgq0HadcqDaDEyOcUor+ik'
    b'wT2QlkeB89IWErmP/rHLCvGgGZm7ZgOtZWPbjhdrDxNkvtZXT86rMIfopME9sBdRv9OAtGuU'
    b'ByJGbRaaXEe/c8RhdA88QocbqXQ+/aIFab69yO3vvIvVMg4k3RnEgspsrQuG8irMCT3KamyW'
    b'3G6QSi5jB3vRMh6g3zniMPZ0exRlbPpC/JrK+KT2dmYVGhLZsIlWPQN4MV0RIKyhYEfLid5h'
    b'cYytcqpRg/QD34MpyExVOr2qO3vKFyNyRtpCQhOR2DXFYL69yG2xrbRyNh2ZUGjUUDD9sqJn'
    b'WPa8IrLSQMReWLgwxq3jmrIiaVyPkD6KRPVWCcfqi3HPitwW28ruwJmg6XcaCr/S6VWuHKGQ'
    b'OcZZ9srpGwYiz9pstRd9kxMy1uQen+/gdzMkEv9O3pS4gMxbtY2KHmcB6f5OJpbh6qnmTtAp'
    b'Q46yvWVT9O8Gfde8SsfhRR9hyXVZaHOLlWle/4k0xToINi0eZOaKHSR2nYPq/7ZbQKVwVqnM'
    b'wLU2S37AriP+YUgDFiKLbFY6s+DCxyUwgwMjS5IoN0hTrMEjRa4gczftZE/yXISDjgqT+6Up'
    b'uswbVQeiU4b0R/Qqe4V5UuYtb03N1aj1AMglNqoJls4k36naOkDCq2I6ZcgnMYyLQXpjWIuk'
    b'qTndFEZRI/PWfahwHvXBc7HkBITXqYz+j9e69mEGfoCqzQOI5EFo2xesoeHlSMta7AYmqH5D'
    b'IvG7MxJZArQlLGr/0GoxB6bLiVPKtO0Pfg57E9AfUtF9oMxcscMAkMjKPajcYbs3kSYNBQt2'
    b'XtAnv7SlX/sDtvMJyd174y4/2phuMhuweyZbL4S/FEXuOx9XSR082XIf2Jx8Bgsz2bj3P/sM'
    b'KLOi69sS8thlGIY8qVOG2l128Skx9MraCjbvfACwd/ZfinuksXnfzMWBCXgM62fYPbwuxTGY'
    b'yed1ak36JTCfkkS/Nbia3skliDqJ2P6QQOKH+79wgAGlsXkLwg0OtRyBoU9oXfAOrR96hMO6'
    b'PkWGTsTUuuAkEoFXQRxOgMst8qt18f1fOSQFWWonP4841CXAJDT5jtYHb9f6qpO1wcPT2H1y'
    b'jk6r7auh6usYFHwNuAP72bDakOepjB6SDL39pxYxv4kmXwKqHOo8DOUaMK5hS80mrWM58Abo'
    b'NrR4Iqrbx3pOIs1PO6mhoaqTMIziHCOrmij9gGGIcSKW9VlEMs0VvZ1k4or2wuDaNaCEV2/U'
    b'abXnYVnPYHfPyCHoAGBC24+LB8PmC2MG4MiAiDEV5XJ39LiN7J89MpuGkghfS3eWXtrbpMxa'
    b'80+QrwCe75nwKVoU0ToJxx5NV6DDcZo0RReCdQGQ9y2LPkWPokyTcLzDqb1OHxSkqfkxRCeA'
    b'bOqsrI9PG3sQrpRIrKmzgraeVCUcX4qYJwIlk5LXxzXWItbpdg8asj1V0pac8RSUG/DHhT7t'
    b's4CyxAkSbn7ebgVHc3XSQEIisZ9jGScg+ieKMMzdxwWE18A6R5piF8ut69rPf5iGjCaLZfaa'
    b'NyUcvwgYTSoKosjn+Hwy5AVUL6Vf7IRMM+dmegAfsDdFP5frdcE6ypiI6Nkg41BsBiX6FB3C'
    b'a6g+jgb+IJHVr2TbXFYG3KcplbhyHjBPGwiwaciRmIERCEdiUQlWb0SKIrVHeiTd9sj0KK8j'
    b'NnMmFxpKEmE7sBlYjcrrqPWyzI6nS6yZEUW/PuEFfkR07vADBnw8xTegj6f4BvTxFN+APp7i'
    b'G9DHU3wD+niKb0AfT/EN6OMpvgF9PMU3oI+n+Ab08RTfgD6e4hvQx1N8A/p4im9AH0/xDejj'
    b'Kb4BfTzFN6CPp/gG9PEU34A+nuIb0MdTfAP6eIpvQB9P8Q3o4ym+AX08xTdgJogRTfObD+i3'
    b'enNetRQ5vgEzoTL6AsJLh7yuent7meB90uPnhskQnV5ViWXchHIq8AHow6yPh2UBSa+1+fj4'
    b'+Pj4+Pj4+Pj4+KTBfwp2iE4aVcZhG2tImtVYVGBqT6A7sIuk7CQguxArRnMs6j8Rd45vwA7Q'
    b'BgzerzmBpJ6G6GkgxwNB7OXWbgHeAV5AeArDelIam9e6KrgI8Q3YDhoaejySuAyVSxEG7/8r'
    b'nH1mB5d/BuE3GNYCaWz2z9/DN+ABaF1wHMiPQU9PvYAirnxGW4CfUtnrNml4vcWF9osG34CA'
    b'1lWPBJkLnJqfDvcZexUwSZpiT+al3wKkSxtQQ8PLMfb8KHX+neTkzBSnElI/MoPKaENXXEfu'
    b'sgbU6VVDSBp/Ak7C+dguh0LaroaiizD0oq42NuySBtRQ7SlgPYww0GstB7GExK7zZe6mnV4L'
    b'yRddLhxLQ9WnItYTBWg+gDGY3R/SiZheC8kXXcqAOjX4OUQWAT281pIW4WwGB2d4LSNfdJlb'
    b'sNYPPRpNvgD0orDfd+oMZkMvlFnxP3usxXUK+Q+RM3TSkX2oSDwPHJ3Pbsn08009mGxAux0j'
    b'kZXbcyursPBi6iH/lCfvJjvzfWSmVCj+k6Avg67BkvcIiJKgAhgMHIvoGSATSK0ROzdiao5w'
    b'EOy5GQhlobvgKfkrYIdHq9pnF8IdqDVXmprfttXvtNq+JK1rEH4A9HbcY+oqaGFax0hj80rH'
    b'9YuEkjagfnvw4bSa/wY5HOfvte3KpQ9jaijTQAKdWj0Yw7h33/Ke8xbulKb41ZnVLXxK+yk4'
    b'UfZTkAFkYj4lAXqdNMUvzCaKRWbH17H78DOBP2ZQXYErdMrQmkz7L3RK1oA6ZWgNlv4ne58q'
    b'HVQFdoP+hzTFb8+FFpm3vJXdAy4BnnVaFSSAmfCvgEWHkfwhQhlOrn6qCppE9QKJxP+WSzky'
    b'b3krpvVVVHfi7EuhIFdoQ2n+rUryTemUIf0Rvg7q7OonIohxba7Nt6/5xua1YPwUZ0MCAWrY'
    b'UpPhGLKwKUkDYpqXAd1AHI799F4JR+e7omkv/XuGgTedV7QuyrmWAqA0DYhc5bCComzGVNfn'
    b'3KTh9RaEmx1W07Z5xZKj5Ayo9UOPBv2kw2qC6I+ksXmLK6IOxuq2AHgP+2NBAYbp9Krh7ony'
    b'hpIzIFbyHIc1FFhLZe873ZDTHhJZuQdhPk6nh5Iy3h1F3lF6BhScGlAQZud9b4bF7zKodWLO'
    b'dXhMSRlQGwignOasEklEfuuSpLRIJPZvoDk19WO71ijXBHlESRmQrVXHIlQ4qiPyrMyKrndJ'
    b'USd98wTi6El9hF5Z6+z9FTilZUA1js+g0sLcC7GLLHZYoYxeks+QMtcpMQNqBgaUF3MvxCaW'
    b'/tNxHbGOdEGJZ5SWAQ2GOq6TTLzqghJ79O+1EpxuxbScv8cCprQCUtUY4mCZVYGtmGat1gVr'
    b'3RPVAVt2ALyLMshBBoaSugKWlgHRKgeFBaiEdpKN5xtH2Wak2jUdHlBat2D4mNcCXMegv9cS'
    b'cknJGFAnjSoDyrzW4TKKcrjXInJJyRiQvjtLan4sDQL+FbAwSe4o91pCnujrtYBcUjoGbLFK'
    b'eoPVfgRKKXVH6Riwe5/dXkvIGwOHl8zsRekYsMeKrpPWLLmtZB62SsaAbckdu0a62/Lybl5L'
    b'yBUlY8A21uN8G2bxYZWVzBWwZMYSbawFHGzilu9KU/SXrqnx6ZRSuwI2Oyx/rCsqfGxTWgYU'
    b'fcthjZKLMC42SsuAlvGyswo6Uq8dVvrrxwVMaRkw4DjA06C8tST32xYLpWXAxthq0E2ogydh'
    b'5WsuKvLphJIyoKSmYBY7PF5rnE6tKako42KipAwIgOJ0o4+JWN9yRYtPp5SeAZPmY+DonF5F'
    b'5Js6rbbWJUU+HVByBpS5azY43O4oQDmW9Qu3NPmkp+QMCIDw+wxqfVnrg1fkXItPh5SmAXe1'
    b'/hllM04zkarOKcUMVIVMSRpQ5q37EGQWjjORSk+Sxr06vaq7W9p8DqQkDQiAKbOBHTiPjhlN'
    b'0nhIG0YUTMhTqeaHhhI2oMxa8z4qM8nsLJQvsHn74/rtwZ7uQNPpVZVaF7yercHnvdThJiW9'
    b'j0JDw8uh5V8ImY3rlGZgskRij+RWWQddNmCwufpzIJcjXAZ0R3hXwrEj8qUhn5S0AQG0rmoC'
    b'GIvI7lT0v2HJz2V29KkcStuHTqsZhDIGS88AnYDsy36Q0uwbsLjR+pq5qF6XfUOsxOCPJOVv'
    b'9Kh4UWau2OGo+qRRZVRsrMUyh2MyDNXjgNM58CDFQ78ovgGLG72ytoLeyedAMkjfdmBTfPSZ'
    b'JYFVQBShGZUPEd1ryL4oFaQOxu4DVADVQBAO2VLZ+ZXZN2Dxo1OGHIVhLkNymlnAyW098yFA'
    b'CRuwZJ+CD0bmrH0L1QmkpmZy1qxLZbsMXcaAADI7/hJqnQd84LUWnxRdyoAAEml+GjU/hxL1'
    b'WotPFzQggERWvwIto0Gebnup9PcSFyhd0oAAEtmwicroWJCpwE7At6EHdFkDAkgDljRF52Ba'
    b'nwC5H8Fq+5VvxTzRpQ24F2lsXitN0UswZCTwe7QgjKjAMmAahvVpD3W4ij810A46ZWgNgcTX'
    b'ULkEGOleRxz8F0iAvoQYD4G5QMKrYq71XSD4BuwEraseiXA+yGdRPkf6DKXtTTR3Nvm8C9Xn'
    b'MWQplv6d5O5lMnfTzlzoLhZ8AzpAQQgFP4EwEmUoBkNRhpLK29wXKAd6khra7CI137gVZCeq'
    b'2xBZiVorMWQlSVnJxmhMFjjaQOXj45NL/h+HkDwWxxeFugAAAABJRU5ErkJggg==')
getLogoData = Logo.GetData
getLogoImage = Logo.GetImage
getLogoBitmap = Logo.GetBitmap
getLogoIcon = Logo.GetIcon




'''
files.py: utilities for working with files.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2019-2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   lxml import etree
import os
from   os import path
import shutil
import subprocess
import sys
import tarfile
import tempfile
import webbrowser
import zipfile
from   zipfile import ZipFile, ZIP_STORED, ZIP_DEFLATED

from .debug import log


# Constants.
# .............................................................................

_EPRINTS2ARCHIVES_REG_PATH = r'Software\Caltech Library\eprints2archives\Settings'


# Main functions.
# .............................................................................

def readable(dest):
    '''Returns True if the given 'dest' is accessible and readable.'''
    return os.access(dest, os.F_OK | os.R_OK)


def writable(dest):
    '''Returns True if the destination is writable.'''

    # Helper function to test if a directory is writable.
    def dir_writable(dir):
        # This is based on the following Stack Overflow answer by user "zak":
        # https://stackoverflow.com/a/25868839/743730
        try:
            testfile = tempfile.TemporaryFile(dir = dir)
            testfile.close()
        except (OSError, IOError) as e:
            return False
        return True

    if path.exists(dest) and not path.isdir(dest):
        # Path is an existing file.
        return os.access(dest, os.F_OK | os.W_OK)
    elif path.isdir(dest):
        # Path itself is an existing directory.  Is it writable?
        return dir_writable(dest)
    else:
        # Path is a file but doesn't exist yet. Can we write to the parent dir?
        return dir_writable(path.dirname(dest))


def file_is_empty(file):
    return os.stat(file).st_size == 0


def file_in_use(file):
    '''Returns True if the given 'file' appears to be in use.  Note: this only
    works on Windows, currently.
    '''
    if not path.exists(file):
        return False
    if sys.platform.startswith('win'):
        # This is a hack, and it really only works for this purpose on Windows.
        try:
            os.rename(file, file)
            return False
        except:
            return True
    return False


def filename_basename(file):
    parts = file.rpartition('.')
    if len(parts) > 1:
        return ''.join(parts[:-1]).rstrip('.')
    else:
        return file


def filename_extension(file):
    parts = file.rpartition('.')
    if len(parts) > 1:
        return '.' + parts[-1].lower()
    else:
        return ''


def module_path():
    '''Returns the absolute path to our module installation directory.'''
    # The path returned by module.__path__ is to the directory containing
    # the __init__.py file.
    this_module = sys.modules[__package__]
    module_path = this_module.__path__[0]
    return path.abspath(module_path)


def datadir_path():
    '''Returns the path to eprints2archives internal data directory.'''
    return path.join(module_path(), 'data')


def desktop_path():
    '''Returns the path to the user's desktop directory.'''
    if sys.platform.startswith('win'):
        return path.join(path.join(os.environ['USERPROFILE']), 'Desktop')
    else:
        return path.join(path.join(path.expanduser('~')), 'Desktop')


def eprints2archives_path():
    '''Returns the path to where eprints2archives is installed.'''
    # The path returned by module.__path__ is to the directory containing
    # the __init__.py file.  What we want here is the path to the installation
    # of the eprints2archives binary.
    if sys.platform.startswith('win'):
        from winreg import OpenKey, CloseKey, QueryValueEx, HKEY_LOCAL_MACHINE, KEY_READ
        try:
            if __debug__: log('reading Windows registry entry')
            key = OpenKey(HKEY_LOCAL_MACHINE, _EPRINTS2ARCHIVES_REG_PATH)
            value, regtype = QueryValueEx(key, 'Path')
            CloseKey(key)
            if __debug__: log(f'path to windows installation: {value}')
            return value
        except WindowsError:
            # Kind of a problem. Punt and return a default value.
            return path.abspath('C:\Program Files\eprints2archives')
    else:
        return path.abspath(path.join(module_path(), '..'))



'''
eprints.py: EPrints-specific utilities

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code is
open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import codecs
from   collections import defaultdict
from   lxml import etree
import os
from   os import path
import shutil

from .data_helpers import parse_datetime
from .debug import log
from .exceptions import *
from .network import net, hostname, scheme, netloc
from .ui import warn, alert


# Constants.
# .............................................................................

_EPRINTS_XMLNS = 'http://eprints.org/ep2/data/2.0'
'''XML namespace used in EPrints XML output.'''


# Main functions.
# .............................................................................

class EPrintServer():

    def __init__(self, api_url, user, password):
        if __debug__: log('creating EPrintsServer object for ', api_url)
        self._api_url    = api_url
        self._protocol   = scheme(self._api_url)
        self._netloc     = netloc(self._api_url)
        self._hostname   = hostname(self._api_url)
        self._user       = user
        self._password   = password
        # List of all record identifiers known to the server:
        self._index      = []
        # Cache of record XML objects obtained from the server:
        self._records    = {}


    def __str__(self):
        return self._hostname


    # Public methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def index(self, as_int = False):
        '''Return a list of all record identifiers from the server.  By
        default, the numbers are returned as strings, not ints, because the
        most common uses of the results end up using strings anyway.  However,
        if the parameter 'as_int' is set to True, the list will be returned
        as integers.'''

        # The raw index returned by EPrints is in XHTML format, like this:
        #
        #   <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
        #       "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        #   <html xmlns="http://www.w3.org/1999/xhtml">
        #   <head>
        #     <title>EPrints REST: Eprints DataSet</title>
        #     <style type="text/css">
        #       body { font-family: sans-serif; }
        #     </style>
        #   </head>
        #   <body>
        #     <h1>EPrints REST: Eprints DataSet</h1>
        #   <ul>
        #   <li><a href='4/'>4/</a></li>
        #   <li><a href='4.xml'>4.xml</a></li>
        #   <li><a href='5/'>5/</a></li>
        #   <li><a href='5.xml'>5.xml</a></li>
        #   ...
        # We extract the identifiers by looping over the XML elements.

        if self._index:
            if __debug__: log('using cached copy of index')
        else:
            xml = etree.fromstring(self._eprints_raw_index())
            if __debug__: log("parsing raw EPrints index to get record id's")
            for node in xml.findall('.//{http://www.w3.org/1999/xhtml}a'):
                if 'href' in node.attrib and node.attrib['href'].endswith('xml'):
                    self._index.append(node.attrib['href'].split('.')[0])
        if self._index and as_int:
            if __debug__: log('returning list of identifiers as ints')
            return [int(x) for x in self._index]
        else:
            return self._index


    def eprint_id_url(self, id_or_record, verify = True):
        '''Return the main URL of the web page for the record on this server.

        This URL is for the page located at "https://.../id/eprint/N", where N
        is the eprint id of the record in question.  Note that EPrints servers
        typically also make the same records available from a URL of the form
        "https://.../N"; this other form is returned by the companion method
        eprint_page_url().

        The argument value for "id_or_record" can be either an identifier or
        an EPrint XML object (as a Python lxml.etree object).

        A server's index may report EPrint record identifiers for records that
        are not actually exposed to the public (e.g., because they've been
        deleted or otherwise marked as not publicly displayed).  Thus, given
        an identifier or even an XML record, it is possible to create URLs
        that do not lead to valid web pages.  Without additional knowledge
        about a specific server's conventions for public/private record
        visibility, the only way to know is to attempt to access the web page
        at the expected address.  Consequently, this method will attempt to do
        a network lookup on the URL, unless the parameter "verify" is given a
        value of False.
        '''
        if isinstance(id_or_record, (str, int)):
            url = f'{self._protocol}://{self._netloc}/id/eprint/{id_or_record}'
        else:
            eprintid = self._xml_field_value(id_or_record, 'eprintid')
            url = f'{self._protocol}://{self._netloc}/id/eprint/{eprintid}'
        if verify:
            (response, error) = net('head', url, timeout = 10)
            if error:
                return None
        return url


    def eprint_page_url(self, id_or_record, verify = True):
        '''Return the web page address for the record on this server.

        This URL is for the page located at "https://.../N", where N is the
        eprint id of the record in question.  Note that EPrints servers
        typically also make the same records available from a URL of the form
        "https://.../id/eprint/N"; this other form is returned by the companion
        method eprint_id_url().

        The argument value for "id_or_record" can be either an identifier or
        an EPrint XML object (as a Python lxml.etree object).

        A server's index may report EPrint record identifiers for records that
        are not actually exposed to the public (e.g., because they've been
        deleted or otherwise marked as not publicly displayed).  Thus, given
        an identifier or even an XML record, it is possible to create URLs
        that do not lead to valid web pages.  Without additional knowledge
        about a specific server's conventions for public/private record
        visibility, the only way to know is to attempt to access the web page
        at the expected address.  Consequently, this method will attempt to do
        a network lookup on the URL, unless the parameter "verify" is given a
        value of False.
        '''
        if isinstance(id_or_record, (str, int)):
            url = f'{self._protocol}://{self._netloc}/{id_or_record}'
        else:
            eprintid = self._xml_field_value(id_or_record, 'eprintid')
            url = f'{self._protocol}://{self._netloc}/{eprintid}'
        if verify:
            (response, error) = net('head', url, timeout = 10)
            if error:
                return None
        return url


    def eprint_xml(self, eprintid):
        '''Return an XML object identified by the given record identifier.'''
        eprintid = str(eprintid)
        if eprintid in self._records:
            if __debug__: log(f'returning cached XML for record {eprintid}')
            return self._records[eprintid]

        if __debug__: log(f'getting XML for {eprintid} from server')
        try:
            response = self._get(f'/eprint/{eprintid}.xml')
        except Exception as ex:
            # Our EPrints server sometimes returns with access forbidden for
            # specific records.  Our caller may simply move on, so we store
            # a value before bubbling up the exception.
            if __debug__: log(f'{str(ex)} for {eprintid}')
            self._records[eprintid] = None
            raise ex
        if response is None:
            self._records[eprintid] = None
            return None
        xml = etree.fromstring(response.content)
        self._records[eprintid] = xml
        return xml


    def eprint_field_value(self, id_or_record, field):
        '''Return the value of 'field' of the record 'id_or_record'.  If
        'id_or_record' is a number (either as a string or an integer), this
        method will ask the server for the field value using the REST API;
        if 'id_or_record' is an XML object, then the value is looked up in the
        object itself.'''

        if isinstance(id_or_record, (str, int)):
            id_or_record = str(id_or_record)
            if field == 'eprintid':
                # Data lookups are unnecessary since we have the id already.
                # This case is here to provide a uniform calling experience.
                return id_or_record
            elif id_or_record in self._records:
                # We have a copy of the XML for this one.  Use it.
                if __debug__: log(f'using cached copy of record {id_or_record}')
                xml = self._records[id_or_record]
            else:
                # Contact the server.
                if __debug__: log(f'{id_or_record} not cached -- asking server')
                field_url = f'/eprint/{id_or_record}/{field}.txt'
                try:
                    response = self._get(field_url)
                except NoContent as ex:
                    if __debug__: log(f'No content for {field} in {id_or_record}')
                    return None
                except AuthenticationFailure as ex:
                    if __debug__: log(f'Auth failure for {field} in {id_or_record}')
                    return None
                except Exception as ex:
                    if __debug__: log(f'{str(ex)} for {field} in {id_or_record}')
                    raise
                if __debug__: log(f'got response: {response.text}')
                return response.text if response and response.text != '' else None
        else:
            xml = id_or_record
        value = self._xml_field_value(xml, field)
        if __debug__: log(f'obtained value: {value}')
        return value


    # Internal methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _get(self, op):
        if not self._api_url:
            return None

        url = self._api_url
        start = url.find('//')
        if start < 0:
            raise BadURL(f'Unable to parse "{url}" as a normal URL')
        if self._user and self._password:
            endpoint = url[:start + 2] + self._user + ':' + self._password + '@' + url[start + 2:] + op
        elif self._user and not self._password:
            endpoint = url[:start + 2] + self._user + '@' + url[start + 2:] + op
        else:
            endpoint = url[:start + 2] + url[start + 2:] + op

        (response, error) = net('get', endpoint)
        if not error and response:
            return response
        else:
            if __debug__: log(f'got {type(error)} error for {url}')
            raise error


    def _eprints_raw_index(self):
        if __debug__: log(f'asking {self._hostname} for index of records')
        response = self._get('/eprint')
        if response and response.text and response.text.startswith('<?xml'):
            return response.content
        else:
            # This shouldn't happen.
            raise InternalError('Internal error processing server response')


    def _xml_field_value(self, xml, field):
        if xml is not None and etree.iselement(xml):
            node = xml.find('.//{' + _EPRINTS_XMLNS + '}' + field)
            return node.text if node != None else ''
        else:
            raise InternalError(f'Not an XML object: {xml}')
